/* knapuser.c includes file bb.skl
 *
 * This fork95 program solves the 0-1 Knapsack problem using the
 * Branch and Bound skeleton prototype specified in file bb.skl
 * For information on Branch and Bound and the Knapsack Problem
 * you can use the book of Horowitz and Sahni,
 * "Fundamentals of Computers Algorithms" (Pitman)
 * For more detailed information about the Knapsack problem, see
 * "Knapsack Problems. Algorithms and Computer Implementations." by
 * Silvano Martello and Paolo Toth. John Wiley and Sons.
 * The  user has to specify the same items than in a sequential 
 * branch and bound algorithm. That is:
 * 1) the type defining a problem. Has to be called "problemt"
 * 2) the type "number" returned for the function to optimize
 * 3) the type "index" of the number of problems in the heap 
 *   (usually an "int" will be sufficient)
 * 4) the function setProblem:
 *              problemt * setProblem( ...) 
 *    this functions return the problem built from the data
 *    constituting the subsequent parameters.
 * 5) the function numChildren: 
 *              index numChildren(problemt *Prob) 
 *    returning the number of children of problem Prob.
 * 6) the function setChild:
 *      child * setChild(problemt *father,index order) 
 *    it returns the "order"-th child problem 
 *    among the children of problem "father".
 * 7) the function
 *      void lowerUpper(problemt Prob, number *L, number *U) 
 *    it returns in L the lower bound and in U the upper 
 *    bound of problem "Prob".
 * 8) The variable "bestSol" of type "number". It will contain
 *    the solution to the problem.
 * The skeletton returns as output the functions
 *
 *			insertProblem(Prob)
 * 	  	branchAndBound()
 * the first one inserts a problem in the heap, the second 
 * solves the problem using the branch and bound technique.
 * Once the functions defining the b&b structure has been provided
 * the user has to insert the problem in the heap and call to 
 * branch and bound function (see main). 
 * Author: Casiano Rodriguez Leon. Email: casiano@ull.es
 * March, 1999
*/

#include <fork.h>
#include <io.h>
#include <stdlib.h>
#include <assert.h>

typedef int index;
typedef int number;
typedef int boolean;

typedef struct {
	number 	k /* object index */, 
					C /* remaining capacity */, 
					P /*accumulated profit */;
} problemt;

#define MAXNUMBEROFOBJECTS 64 
#define  INFINITY 8192
/* INFINITY has to be set to a value greater 
   than the capacity M of the Knapsack */

#define ns "%d\n"

/* knapsack support variables */
sh number N, M;
sh number 
		w[MAXNUMBEROFOBJECTS]  /* weights */, 
		p[MAXNUMBEROFOBJECTS]; /* profits */
sh number bestSol = -INFINITY; 

void readKnap(void) {
#define STRIDE 1

	FILE *in;
  index i;

	if ((in = fopen("knap.dat","r+")) == NULL) {
	 fprintf(stderr, "Cannot open input file knap.dat.\n");
	 fprintf(stderr,"Input file format is:\nNumber of objects\nCapacity\nweight 0\nprofit 0\n...\n");
	 exit(1);
	}
	fscanf(in,ns,&N);
	assert(N < MAXNUMBEROFOBJECTS);
	fscanf(in,ns,&M);
	pprintf("N = %d\n",N);
	pprintf("M = %d\n",M);
	assert(INFINITY > M);
	for(i=0;i<N;i++) {
		fscanf(in,ns,&w[i]);
		fscanf(in,ns,&p[i]);
		if (i%STRIDE == 0) {
			pprintf("w = %d\n",w[i]);
			pprintf("p = %d\n",p[i]);
	 }
	}
	w[N] = INFINITY; p[N] = 0; /* sentinel object */
	fclose(in);
}

void lowerUpper(problemt *Prob, number *L, number *U) {
  /* naive approach */
  number i, weig = 0, prof = Prob->P;

  if (Prob->C < 0) {
	 *L = -INFINITY; *U = -INFINITY;
  }
  else {
	 for(i=Prob->k; weig <= Prob->C; i++) {
		weig += w[i]; prof += p[i];
	 }
	 /* invariant:  weig > Prob->C */
	 i--;
	 weig -= w[i];
	 prof -= p[i];
	 *L = prof;  /* see Martello and Toth pag 16. Th 2.1 */
	 *U = prof+(p[i]*(Prob->C-weig))/w[i];
  }
}

problemt* setProblem(number k, number C, number P) {
	problemt *Prob; 

  Prob = (problemt *) shmalloc(sizeof(problemt)); 

  Prob->k = k;
  Prob->C = C;
  Prob->P = P;
	pprintf("setting problem %d %d %d\n",Prob->k,Prob->C,Prob->P);

	return Prob;
}

index numChildren(problemt *Prob) {
  if (Prob->k < N) return 2; else return 0;
}

problemt * setChild(problemt *father,index order) {

  return setProblem(father->k+1,
				 father->C-(1-order)*w[father->k],
				 father->P+(1-order)*p[father->k]
				);
}

printData(problemt Prob) {

	pprintf("(k = %d, C = %d, P = %d)\n",Prob.k,Prob.C,Prob.P);
}
/** The b&b skeleton must be included after the  definition of **/
/********* the b&b instance functions ***************************/

#include "bb.skl"

int main(void) {
  problemt *Prob;

	if ($ == 0) {
		readKnap();
		Prob = setProblem(0, M, 0);
	}

	pprintf("calling b&b\n...");
	
	branchAndBound(Prob);
	
	pprintf("****** best solution = %d *******\n",bestSol);
	barrier;
}
